
/*
 * $Id: HoCoDistr.cc,v 1.1.2.1 1999/02/16 03:43:32 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include "HoCoDistr.h"

// XXX: hard coded parameters for repeat inter-request distances
static const double TheRepDistAvg = 20;
static const double TheRepDistDev = 10;
// best oid to repeat is searched in [ offset-delta, offset+delta]
static const int TheSearchDelta = 1; // XXX: hard coded

HoCoDistr::HoCoDistr(RndGen *aGen, double aHitR, double aHotR, double aColdR, int aHotSize):
	OidDistr(aGen, aHitR), theHotRatio(aHotR), theColdRatio(aColdR),
	theHotSize(aHotSize),
	theRepDistance(0), theSearchDelta(0),
	theLastHotOid(0), theLastWarmOid(0), theLastColdOid(1) {

	theNotColdRatio = 1 - theColdRatio;

	theMaxHotOid = 2*theHotSize;
	theHistory.capacity(500000);
	theRepDistance = LognDistr::ViaMean(theGen, TheRepDistAvg, TheRepDistDev);
	theSearchDelta = TheSearchDelta;
}

int HoCoDistr::genOid() {
	const double sw = theGen->trial();

	// hot/warm hit or warm/hit miss?
	const int oid = sw < theHitRatio && theLastWarmOid ?
		(sw < theHotRatio || theLastWarmOid <= theMaxHotOid ?
			genHotHit() : genWarmHit()) :
		(sw <= theNotColdRatio || theLastHotOid < theMaxHotOid ?
			genWarmMiss() : genColdMiss());
	Assert(oid > 0);

	theHistory.insert(oid);
	return oid;
}

int HoCoDistr::genColdMiss() {
	return theLastColdOid += 2;
}

int HoCoDistr::genWarmMiss() {
	if (theLastHotOid < theMaxHotOid)
		theLastHotOid += 2;
	return theLastWarmOid += 2;
}

int HoCoDistr::genHotHit() {
	return repeat(2, theLastHotOid);
}

int HoCoDistr::genWarmHit() {
	return repeat(theMaxHotOid+2, theLastWarmOid);
}

int HoCoDistr::repeat(int firstOid, int lastOid) {
	const int lastHistOff = theHistory.depth() - 1;

	// temporal locality: distance between two req for one obj
	// XXX: offset must be scaled if more than one polyclt is used?
	const int offset = Min(lastHistOff, (int) theRepDistance->trial());

	// treat offset above as an approximate location
	const int beg = Max(0, offset-theSearchDelta);
	const int end = Min(lastHistOff, offset+theSearchDelta);

	// find smallest hit within area subject to [firstOid, lastOid]
	int best = 0;
	for (int i = beg; i <= end; i++) {
		const int oid = theHistory.at(i);
		if (!(oid & 1) && firstOid <= oid && oid <= lastOid && (best > oid || !best))
			best = oid;
	}

	if (best)
		return best;

	// generate zipf-like distributed hit
	const int win = lastOid - firstOid;
	return ~1 & (firstOid + (int)floor(pow(win, theGen->trial())));
}

ostream &HoCoDistr::print(ostream &os) const {
	return os << pdfName() << "(" 
		<< theHitRatio << ", "
		<< theHotRatio << ","
		<< theColdRatio << ", "
		<< theHotSize << ")";
}
